@@ -3,11 +3,11 @@ from .utils import *
 # from utils import *
 	
 class sgRNA:
-	def __init__(self,chr=None,start=None,end=None,seq=None,sgRNA_name=None,strand=None,cut_position=None,mutation_pos = None,mutation_ref = None,mutation_alt = None,variant_id=None,dist_dict=None,opposite_strand_sgRNAs=None,all_sgRNA_df=None,target_fa=None,scaffold_seq=None,user_target_pos=None,user_ref=None,user_alt=None,is_dPAM=None,target_to_sgRNA=None,**kwargs):
+	def __init__(self,chr=None,start=None,end=None,seq=None,sgRNA_name=None,strand=None,cut_position=None,mutation_pos = None,mutation_ref = None,mutation_alt = None,variant_id=None,dist_dict=None,opposite_strand_sgRNAs=None,all_sgRNA_df=None,target_fa=None,scaffold_seq=None,user_target_pos=None,user_ref=None,user_alt=None,offset=None,target_to_sgRNA=None,PAM=None,DeepSpCas9=None,**kwargs):
 	
 		"""
 		
-		search key is chr_start_end_seq because seq can be duplicated, this is unique
+		search key is chr_start_end_seq because seq can be duplicated, this chr_start_end_seq is unique. This name doesn't show strand, but seq is alwasy 5-3 direction
 		
 		
 		Searching steps
@@ -40,7 +40,10 @@ class sgRNA:
 		self.seq = seq
 		self.variant_id = variant_id
 		self.target_to_sgRNA = target_to_sgRNA
-		self.is_dPAM = is_dPAM
+		self.DeepSpCas9 = DeepSpCas9
+		self.is_dPAM = 0
+		self.PAM = PAM
+		self.offset = offset
 		self.sgRNA_name = sgRNA_name
 		# print ("init",sgRNA_name)
 		self.uid = str(uuid.uuid4()).split("-")[-1]
@@ -72,7 +75,8 @@ class sgRNA:
 		# self.RTT_feature_list = []
 		# self.ngRNA_feature_list = []
 		self.PBS_feature_list = ["PBS_GC",'PBS_length']
-		self.RTT_feature_list = ["target_to_RTT5","RTT_GC","RTT_length","0","1","2","3","4","5","6","7","8","9","10","11","12","13"]
+		# self.RTT_feature_list = ["target_to_RTT5","RTT_GC","RTT_length","0","1","2","3","4","5","6","7","8","9","10","11","12","13"]
+		self.RTT_feature_list = ["target_to_RTT5","RTT_GC","RTT_length","0","1","2","3","4","5","6","7","8","9"]
 		self.ngRNA_feature_list = ['sgRNA_distance_to_ngRNA','is_PE3b']
 		
 		self.rawX = pd.DataFrame()
@@ -116,7 +120,7 @@ class sgRNA:
 		self.PBS_df.columns = ['chr','start','end']
 		self.PBS_df["strand"] = get_opposite_strand(self.strand)
 		self.PBS_df.index = ["%s_PBS_%s"%(self.uid,i) for i in range(self.PBS_df.shape[0])]
-		temp = get_fasta_simple(self.target_fa,self.PBS_df, self.target_pos)
+		temp = get_fasta_simple(self.target_fa,self.PBS_df, self.user_target_pos)
 		self.PBS_df['seq'] = temp[3].tolist()
 		
 		if self.strand == "+": ## when sgRNA is positive strand, RTT should use the negative strand
@@ -126,7 +130,133 @@ class sgRNA:
 		self.PBS_df['PBS_length'] = [len(x) for x in self.PBS_df['seq'] ]
 		
 
-	def find_RTT(self,min_RTT_length=7,max_RTT_length=40,min_distance_RTT5=5,**kwargs):
+	def find_RTT(self,debug=0,min_RTT_length=10,max_RTT_length=20,max_max_RTT_length=40,min_distance_RTT5=5,**kwargs):
+		"""find RTT sequences as bed format df
+		
+		Output
+		--------
+		
+		RTT_df
+		
+		RTT_feature_list
+		
+		"""
+		out = []
+		chr = self.chr
+		pbs_start = None
+		pbs_end = None
+		user_max_RTT_length = max_RTT_length
+		large_deletion_flag=False
+		if len(self.ref)+min_distance_RTT5 > max_RTT_length: ## in case of large deletion
+			large_deletion_flag = True
+			max_RTT_length = max_RTT_length+len(self.ref)
+		# target_to_RTT5_feature=[]
+		deletion_length = len(self.ref)-len(self.alt)
+		if self.strand=="+":
+			start = self.cut_position # remember out cut position, the actual nucleotide, we use -4
+			pbs_end = start
+			pbs_start = pbs_end - 14 
+			for l in range(min_RTT_length,max_RTT_length+1+max(0,deletion_length)):
+				end = start+l
+				if start+1<=self.target_pos <=end-min_distance_RTT5:
+					out.append([chr,start,end,end-self.target_pos-max(0,deletion_length)])
+		if self.strand=="-":
+			end = self.cut_position - 1
+			pbs_start = end
+			pbs_end = pbs_start + 14 
+			for l in range(min_RTT_length,max_RTT_length+1+max(0,deletion_length)):
+				start = end-l
+				if end>=self.target_pos >=start+1+min_distance_RTT5:
+					out.append([chr,start,end,self.target_pos-start-1-max(0,deletion_length)])	
+					
+		current_max_RTT_length = max_RTT_length+5
+		while len(out) == 0:
+			if current_max_RTT_length > max_max_RTT_length:
+				break
+			out,_,no_RTT_flag = self.find_longer_RTT(min_RTT_length=min_RTT_length,max_RTT_length=current_max_RTT_length,min_distance_RTT5=min_distance_RTT5,**kwargs)
+			if len(out) > 0:
+				print ("max_RTT_length increased from %s to %s"%(max_RTT_length,current_max_RTT_length))
+				break
+			current_max_RTT_length += 5
+					
+		if len(out) == 0:
+			## valid RTT not found
+			self.no_RTT = True
+			print ("No valid RTT found given current max length:%s"%(max_max_RTT_length))
+			return 0
+		self.RTT_df = pd.DataFrame(out)
+		self.RTT_df.columns = ['chr','start','end','target_to_RTT5']
+		self.RTT_df["strand"] = get_opposite_strand(self.strand)
+		self.RTT_df.index = ["%s_RTT_%s"%(self.uid,i) for i in range(self.RTT_df.shape[0])]
+		if debug>10:
+			print (self.target_fa)
+			print (self.user_target_pos)
+		temp = get_fasta_simple(self.target_fa,self.RTT_df, self.user_target_pos)
+		self.RTT_df['old_seq'] = temp[3].tolist()
+	
+		## add variant
+		# relative_pos = self.target_pos-r['start']-1 # start is 0-index
+		if debug>10:
+			pd.set_option('display.max_columns', None)
+			print (self.RTT_df)
+		self.RTT_df['seq'] = [self.add_variant(r['old_seq'],self.target_pos-r['start']-1,self.ref,self.alt) for i,r in self.RTT_df.iterrows()]
+		self.RTT_df = self.RTT_df[self.RTT_df['seq']!=0]
+		if self.strand == "+": ## when sgRNA is positive strand, RTT should use the negative strand
+			self.RTT_df['seq'] = [revcomp(x) for x in self.RTT_df['seq']]
+		# print (self.RTT_df)
+		self.RTT_df['RTT_length'] = [len(x) for x in self.RTT_df['seq'] ]
+		if large_deletion_flag:
+			self.RTT_df = self.RTT_df[self.RTT_df['RTT_length']<=user_max_RTT_length]
+		self.RTT_df = self.RTT_df[self.RTT_df['RTT_length']>=min_RTT_length]
+		# filter out first C
+		self.RTT_df['firstC'] = [x[0] for x in self.RTT_df['seq']]
+		self.RTT_df = self.RTT_df[self.RTT_df['firstC']!="C"]
+		self.RTT_df = self.RTT_df.drop(['firstC'],axis=1)	
+
+		
+		current_max_RTT_length = max_RTT_length+5
+		while self.RTT_df.shape[0] == 0:
+			if debug>=10:
+				print ("increasing max_RTT_length to:", current_max_RTT_length)
+			if current_max_RTT_length > max_max_RTT_length:
+				break
+			_,RTT_df,no_RTT_flag = self.find_longer_RTT(min_RTT_length=min_RTT_length,max_RTT_length=current_max_RTT_length,min_distance_RTT5=min_distance_RTT5,**kwargs)
+			if RTT_df.shape[0] > 0:
+				print ("max_RTT_length increased from %s to %s"%(max_RTT_length,current_max_RTT_length))
+				self.RTT_df = RTT_df
+				break
+			current_max_RTT_length += 5		
+
+		
+		
+		if self.RTT_df.shape[0] == 0:
+			## valid RTT not found
+			self.no_RTT = True
+			print ("NO RTT found: new RTT sequence length is longer than: %s"%(max_max_RTT_length))
+			return 0
+
+		
+		## make features
+		## modify it to fit fasta input
+		attached_minimal_PBS = sub_fasta_single(self.target_fa,self.user_target_pos, pbs_start,pbs_end)
+		if self.strand == "+": ## when sgRNA is positive strand, RTT should use the negative strand
+			attached_minimal_PBS = revcomp(attached_minimal_PBS)
+		# print ("attached_minimal_PBS",attached_minimal_PBS)
+		# print (self.RTT_df)
+
+		
+		self.RTT_df['RNAfold_seq']= [(self.scaffold_seq+x+attached_minimal_PBS).replace("T","U") for x in self.RTT_df['seq']]
+		RNAfold_features_df = pd.DataFrame([call_RNAplfold(x,len(self.scaffold_seq))for x in self.RTT_df['RNAfold_seq']])
+		# print (RNAfold_features_df)
+		RNAfold_features_df.index = self.RTT_df.index.tolist()
+		self.RTT_df = pd.concat([self.RTT_df,RNAfold_features_df],axis=1)
+		self.RTT_df['RTT_GC'] = [GC_content(x) for x in self.RTT_df['seq'] ]
+		
+		self.RTT_df.columns = [str(x) for x in self.RTT_df.columns]
+		self.is_dPAM = is_dPAM(self.PAM, self.RTT_df['seq'][0], self.offset)
+		# print (self.RTT_df)
+		
+	def find_longer_RTT(self,min_RTT_length=10,max_RTT_length=20,min_distance_RTT5=5,**kwargs):
 		"""find RTT sequences as bed format df
 		
 		Output
@@ -164,9 +294,7 @@ class sgRNA:
 				if end>=self.target_pos >=start+1+min_distance_RTT5:
 					out.append([chr,start,end,self.target_pos-start-1])	
 		if len(out) == 0:
-			## valid RTT not found
-			self.no_RTT = True
-			return 0
+			return out,self.RTT_df,True
 		self.RTT_df = pd.DataFrame(out)
 		self.RTT_df.columns = ['chr','start','end','target_to_RTT5']
 		self.RTT_df["strand"] = get_opposite_strand(self.strand)
@@ -185,28 +313,14 @@ class sgRNA:
 		if large_deletion_flag:
 			self.RTT_df = self.RTT_df[self.RTT_df['RTT_length']<=user_max_RTT_length]
 		self.RTT_df = self.RTT_df[self.RTT_df['RTT_length']>=min_RTT_length]
+		# filter out first C
+		self.RTT_df['firstC'] = [x[0] for x in self.RTT_df['seq']]
+		self.RTT_df = self.RTT_df[self.RTT_df['firstC']!="C"]
+		self.RTT_df = self.RTT_df.drop(['firstC'],axis=1)	
 		if self.RTT_df.shape[0] == 0:
-			## valid RTT not found
-			self.no_RTT = True
-			return 0
-		## make features
-		attached_minimal_PBS = sub_fasta_single(self.target_fa,self.user_target_pos, pbs_start,pbs_end)
-		if self.strand == "+": ## when sgRNA is positive strand, RTT should use the negative strand
-			attached_minimal_PBS = revcomp(attached_minimal_PBS)
-		# print ("attached_minimal_PBS",attached_minimal_PBS)
-		# print (self.RTT_df)
-
-		
-		self.RTT_df['RNAfold_seq']= [(self.scaffold_seq+x+attached_minimal_PBS).replace("T","U") for x in self.RTT_df['seq']]
-		RNAfold_features_df = pd.DataFrame([call_RNAplfold(x,len(self.scaffold_seq))for x in self.RTT_df['RNAfold_seq']])
-		# print (RNAfold_features_df)
-		RNAfold_features_df.index = self.RTT_df.index.tolist()
-		self.RTT_df = pd.concat([self.RTT_df,RNAfold_features_df],axis=1)
-		self.RTT_df['RTT_GC'] = [GC_content(x) for x in self.RTT_df['seq'] ]
-		
-		self.RTT_df.columns = [str(x) for x in self.RTT_df.columns]
-		
-		# print (self.RTT_df)
+			return out,self.RTT_df,True
+		return out,self.RTT_df,False
+			
 
 	def make_PE3b_ngRNA(self,myIndex,start,end,seq,strand):
 		"""check if target mutation overlaps with ngRNA and update its sequence
@@ -253,7 +367,7 @@ class sgRNA:
 			new_seq = "".join(new_seq)
 			return [myIndex,new_seq,1]
 		
-	def find_nick_gRNA(self,max_ngRNA_distance=150,debug=0,**kwargs):
+	def find_nick_gRNA(self,max_ngRNA_distance=100,max_max_ngRNA_distance=200,debug=0,**kwargs):
 		"""find all valid ngRNAs given the sgRNA name
 		
 		Input
@@ -269,6 +383,7 @@ class sgRNA:
 		# print (self.ngRNA_df)
 		if self.ngRNA_df.shape[0] == 0:
 			self.no_ngRNA = True
+			print ("no ngRNA for %s"%(self.sgRNA_name))
 			return 0
 		self.ngRNA_df.index = ["%s_ngRNA_%s"%(self.uid,i) for i in range(self.ngRNA_df.shape[0])]
 		self.ngRNA_df.columns = ['chr','start','end','seq','sgRNA_name','strand']
@@ -276,11 +391,25 @@ class sgRNA:
 			self.ngRNA_df['sgRNA_distance_to_ngRNA'] = [-self.dist_dict[x][self.sgRNA_name] for x in self.ngRNA_df["sgRNA_name"]]
 		if self.strand == "+":
 			self.ngRNA_df['sgRNA_distance_to_ngRNA'] = [self.dist_dict[x][self.sgRNA_name] for x in self.ngRNA_df['sgRNA_name']]
-		self.ngRNA_df = self.ngRNA_df[self.ngRNA_df['sgRNA_distance_to_ngRNA'].abs()<=max_ngRNA_distance]
+		# print (max_ngRNA_distance,"max_ngRNA_distance")
+		current_ngRNA_df = self.ngRNA_df[self.ngRNA_df['sgRNA_distance_to_ngRNA'].abs()<=max_ngRNA_distance]
+		current_max_ngRNA_distance = max_ngRNA_distance + 20
+		while current_ngRNA_df.shape[0] == 0:
+			if current_max_ngRNA_distance > max_max_ngRNA_distance:
+				self.ngRNA_df = current_ngRNA_df
+				break
+			current_ngRNA_df = self.ngRNA_df[self.ngRNA_df['sgRNA_distance_to_ngRNA'].abs()<=current_max_ngRNA_distance]
+			if current_ngRNA_df.shape[0] > 0:
+				self.ngRNA_df = current_ngRNA_df
+				print ("max ngRNA distance increased from %s to %s:"%(max_ngRNA_distance,current_max_ngRNA_distance))
+				break
+			current_max_ngRNA_distance += 20
+		self.ngRNA_df = current_ngRNA_df
 		if self.ngRNA_df.shape[0] == 0:
+			print ("no ngRNA for max distance: %s"%(max_ngRNA_distance))
 			self.no_ngRNA = True
 			return 0
-		# print (self.ngRNA_df)
+		
 		self.ngRNA_df['is_PE3b'] = 0
 		
 		if len(self.ref) == len(self.alt): # check PE3b
@@ -360,11 +489,17 @@ class sgRNA:
 		
 		if self.no_RTT:
 			return 0
-		## allow PE2 cases
+		## not allow PE2 cases
 		if self.no_ngRNA:
+			print ("no_ngRNA found",self.sgRNA_name)
+			return 0
 			self.ngRNA_df=pd.DataFrame([np.nan]*(len(self.ngRNA_feature_list)+len(rawX_columns))).T
 			self.ngRNA_df.columns = rawX_columns + self.ngRNA_feature_list
-
+		if debug>=10:
+			print (self.variant_id,"showing PBS_df, RTT_df, and ngRNA df")
+			print (self.PBS_df.head())
+			print (self.RTT_df.head())
+			print (self.ngRNA_df.head())
 		X_index = []
 		PBS_selected_rows =[]
 		RTT_selected_rows =[]
@@ -375,11 +510,20 @@ class sgRNA:
 		for s in temp:
 			count += 1
 			current_index = "%s_%s_candidate_%s"%(self.variant_id,self.sgRNA_name,count)
+			# current_index = "%s_%s"%(self.variant_id,count) # shorter name
 			if count == 1  and debug>10:
 				print (current_index)
 			PBS_selected_rows.append(s[0])
 			RTT_selected_rows.append(s[1])
 			ngRNA_selected_rows.append(s[2])
+			if self.ngRNA_df.at[s[2],'is_PE3b']==1:
+				current_index+="_PE3b"
+				# print (s[2],"PE3B")
+			if self.is_dPAM:
+				current_index+="_dPAM"	
+				# print (self.sgRNA_name,"dPAM")		
+			if self.no_ngRNA:
+				current_index+="_PE2"	
 			X_index.append(current_index)
 			
 			
@@ -415,7 +559,8 @@ class sgRNA:
 		X_ngRNA = self.ngRNA_df.loc[ngRNA_selected_rows][self.ngRNA_feature_list]
 		X_ngRNA = X_ngRNA.reset_index(drop=True)
 		self.X = pd.concat([X_PBS,X_RTT,X_ngRNA],axis=1)
-		self.X.index = X_index
+		self.X.index = X_index	
 		self.X['is_dPAM'] = self.is_dPAM
 		self.X['target_to_sgRNA'] = self.target_to_sgRNA
+		self.X['DeepSpCas9'] = self.DeepSpCas9
 
